@@ -8,6 +8,7 @@
 ##############################################################################
 import sys, os, socket
 import xdrlib, time, datetime, decimal, struct
+import itertools
 from firebirdsql.fberrmsgs import messages
 from firebirdsql import (DatabaseError, InternalError, OperationalError, 
     ProgrammingError, IntegrityError, DataError, NotSupportedError,)
@@ -18,6 +19,13 @@ from firebirdsql.wireprotocol import (WireProtocol,
 
 PYTHON_MAJOR_VER = sys.version_info[0]
 
+if sys.version_info < (2, 6):
+    from UserDict import DictMixin as Mapping
+    HAS_MAPPING = False
+else:
+    from collections import Mapping
+    HAS_MAPPING = True
+
 def b2i(b):
     "byte to int"
     if PYTHON_MAJOR_VER == 3:
@@ -26,15 +34,17 @@ def b2i(b):
         return ord(b)
 
 if PYTHON_MAJOR_VER == 2:
-    def bytes(byte_array):
-        return ''.join([chr(c) for c in byte_array])
+    __metaclass__ = type
 
-__version__ = '0.6.2'
+    def bytes(byte_array):
+        return ''.join(chr(c) for c in byte_array)
+
+__version__ = '0.6.3'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
 
-transaction_parameter_block = [
+transaction_parameter_block = (
     # ISOLATION_LEVEL_READ_UNCOMMITTED
     bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_rec_version]),
     # ISOLATION_LEVEL_READ_COMMITED
@@ -45,7 +55,7 @@ transaction_parameter_block = [
     bytes([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_consistency]),
     # ISOLATION_LEVEL_READ_COMMITED_READ_ONLY
     bytes([isc_tpb_version3, isc_tpb_read, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
-]
+)
 
 def Date(year, month, day):
     return datetime.date(year, month, day)
@@ -97,6 +107,21 @@ def convert_timestamp(v):   # Convert datetime.datetime to BLR format timestamp
 
 #------------------------------------------------------------------------------
 class XSQLVAR:
+    type_length = {
+        SQL_TYPE_VARYING: -1,
+        SQL_TYPE_SHORT: 4,
+        SQL_TYPE_LONG: 4,
+        SQL_TYPE_FLOAT: 4,
+        SQL_TYPE_TIME: 4,
+        SQL_TYPE_DATE: 4,
+        SQL_TYPE_DOUBLE: 8,
+        SQL_TYPE_TIMESTAMP: 8,
+        SQL_TYPE_BLOB: 8, 
+        SQL_TYPE_ARRAY: 8, 
+        SQL_TYPE_QUAD: 8,
+        SQL_TYPE_INT64: 8
+        }
+    
     def __init__(self, bytes_to_str):
         self.bytes_to_str = bytes_to_str
         self.sqltype = None
@@ -110,16 +135,11 @@ class XSQLVAR:
         self.aliasname = ''
 
     def io_length(self):
-        if self.sqltype == SQL_TYPE_TEXT:
+        sqltype = self.sqltype
+        if sqltype == SQL_TYPE_TEXT:
             return self.sqllen
-        elif self.sqltype == SQL_TYPE_VARYING:
-            return -1   # First 4 bytes 
-        elif self.sqltype in (SQL_TYPE_SHORT, SQL_TYPE_LONG, SQL_TYPE_FLOAT,
-                SQL_TYPE_TIME, SQL_TYPE_DATE):
-            return 4
-        elif self.sqltype in (SQL_TYPE_DOUBLE, SQL_TYPE_TIMESTAMP,
-                SQL_TYPE_BLOB, SQL_TYPE_ARRAY, SQL_TYPE_QUAD, SQL_TYPE_INT64):
-            return 8
+        else:
+            return self.type_length[sqltype]
 
     def __str__(self):
         s  = '[' + str(self.sqltype) + ',' + str(self.sqlscale) + ',' \
@@ -196,44 +216,43 @@ class XSQLVAR:
         else:
             return raw_value
 
+
+sqltype2blr = {
+    SQL_TYPE_DOUBLE: [27],
+    SQL_TYPE_FLOAT: [10],
+    SQL_TYPE_D_FLOAT: [11],
+    SQL_TYPE_DATE: [12],
+    SQL_TYPE_TIME: [13],
+    SQL_TYPE_TIMESTAMP: [35],
+    SQL_TYPE_BLOB: [9, 0],
+    SQL_TYPE_ARRAY: [9, 0],
+    }
+
 def calc_blr(xsqlda):
     "Calculate  BLR from XSQLVAR array."
     ln = len(xsqlda) *2
     blr = [5, 2, 4, 0, ln & 255, ln >> 8]
     for x in xsqlda:
-        if x.sqltype == SQL_TYPE_VARYING:
+        sqltype = x.sqltype
+        if sqltype == SQL_TYPE_VARYING:
             blr += [37, x.sqllen & 255, x.sqllen >> 8]
-        elif x.sqltype == SQL_TYPE_TEXT:
+        elif sqltype == SQL_TYPE_TEXT:
             blr += [14, x.sqllen & 255, x.sqllen >> 8]
-        elif x.sqltype == SQL_TYPE_DOUBLE:
-            blr += [27]
-        elif x.sqltype == SQL_TYPE_FLOAT:
-            blr += [10]
-        elif x.sqltype == SQL_TYPE_D_FLOAT:
-            blr += [11]
-        elif x.sqltype == SQL_TYPE_DATE:
-            blr += [12]
-        elif x.sqltype == SQL_TYPE_TIME:
-            blr += [13]
-        elif x.sqltype == SQL_TYPE_TIMESTAMP:
-            blr += [35]
-        elif x.sqltype == SQL_TYPE_BLOB:
-            blr += [9, 0]
-        elif x.sqltype == SQL_TYPE_ARRAY:
-            blr += [9, 0]
-        elif x.sqltype == SQL_TYPE_LONG:
+        elif sqltype == SQL_TYPE_LONG:
             blr += [8, x.sqlscale]
-        elif x.sqltype == SQL_TYPE_SHORT:
+        elif sqltype == SQL_TYPE_SHORT:
             blr += [7, x.sqlscale]
-        elif x.sqltype == SQL_TYPE_INT64:
+        elif sqltype == SQL_TYPE_INT64:
             blr += [16, x.sqlscale]
-        elif x.sqltype == SQL_TYPE_QUAD:
+        elif sqltype == SQL_TYPE_QUAD:
             blr += [9, x.sqlscale]
+        else:
+            blr += sqltype2blr[sqltype]
         blr += [7, 0]   # [blr_short, 0]
     blr += [255, 76]    # [blr_end, blr_eoc]
 
     # x.sqlscale value shoud be negative, so b convert to range(0, 256)
-    return bytes([256 + b if b < 0 else b for b in blr])
+    return bytes(256 + b if b < 0 else b for b in blr)
 
 def parse_select_items(buf, xsqlda, connection):
     index = 0
@@ -340,9 +359,6 @@ class PreparedStatement:
         self.statement_type = bytes_to_int(buf[i+3:i+7])
         self._xsqlda = parse_xsqlda(buf[i:], connection, self.stmt_handle)
 
-        # TODO: implement later
-        self.n_input_params = None
-
     def __getattr__(self, attrname):
         if attrname == 'description':
             if len(self._xsqlda) == 0:
@@ -470,7 +486,7 @@ class Cursor:
         self.stmt_handle = h
 
         raise StopIteration()
-            
+
     def fetchone(self):
         try:
             if PYTHON_MAJOR_VER==3:
@@ -480,21 +496,37 @@ class Cursor:
         except StopIteration:
             return None
 
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        r = self.fetchone()
+        if not r:
+            raise StopIteration()
+        return r
+
+    def next(self):
+        return self.__next__()
+
     def fetchall(self):
         return list(self._fetch_records)
 
     def fetchmany(self, size=None):
-        rows = []
         if not size:
             size = self.arraysize
-        r = self.fetchone()
-        while r:
-            rows.append(r)
-            size -= 1
-            if size == 0:
-                break
-            r = self.fetchone()
-        return rows
+        return list(itertools.islice(self._fetch_records, size))
+    
+    # kinterbasdb extended API
+    def fetchonemap(self):
+        return RowMapping(self.fetchone(), self.description)
+    
+    def fetchallmap(self):
+        desc = self.description
+        return [RowMapping(row, desc) for row in self._fetch_records]
+    
+    def fetchmanymap(self, size=None):
+        desc = self.description
+        return [RowMapping(row, desc) for row in self.fetchmany(size)]
 
     def close(self):
         if not hasattr(self, "stmt_handle"):
@@ -503,7 +535,7 @@ class Cursor:
         (h, oid, buf) = self.transaction.connection._op_response()
         delattr(self, "stmt_handle")
 
-    def nextset():
+    def nextset(self):
         raise NotSupportedError()
 
     def setinputsizes(self, sizes):
@@ -511,17 +543,18 @@ class Cursor:
 
     def setoutputsize(self, size, column):
         pass
+    
+    @property
+    def description(self):
+        return [(x.aliasname, x.sqltype, None, x.io_length(), None, 
+                 x.sqlscale, True if x.null_ok else False)
+                for x in self._xsqlda]
+        
+    @property
+    def rowcount(self):
+        # to be implemented
+        return -1
 
-    def __getattr__(self, attrname):
-        if attrname == 'description':
-            r = []
-            for x in self._xsqlda:
-                r.append((x.aliasname, x.sqltype, None, x.io_length(), None, 
-                        x.sqlscale, True if x.null_ok else False))
-            return r
-        elif attrname == 'rowcount':
-            return -1
-        raise AttributeError
 
 class Connection(WireProtocol):
     def uid(self):
@@ -531,8 +564,6 @@ class Connection(WireProtocol):
         else:
             user = os.environ['USER']
             hostname = os.environ.get('HOSTNAME', '')
-#        user = self.str_to_bytes(user)
-#        hostname = self.str_to_bytes(hostname)
         return bytes([1] + [len(user)] + [ord(c) for c in user] 
                 + [4] + [len(hostname)] + [ord(c) for c in hostname] + [6, 0])
 
@@ -592,7 +623,7 @@ class Connection(WireProtocol):
         return c
 
     def begin(self):
-        if not hasattr(self, "db_handle"):
+        if self.closed:
             raise InternalError
         trans = Transaction(self)
         trans.begin()
@@ -606,7 +637,7 @@ class Connection(WireProtocol):
         if len(self._transactions):
             return self._transactions[0]
         return None
-    
+
     def commit(self, retaining=False):
         if self.main_transaction:
             self.main_transaction.commit(retaining=retaining)
@@ -615,7 +646,8 @@ class Connection(WireProtocol):
         return self.main_transaction.savepoint(name)
 
     def rollback(self, retaining=False, savepoint=None):
-        self.main_transaction.rollback(retaining=retaining, savepoint=savepoint)
+        if self.main_transaction:
+            self.main_transaction.rollback(retaining=retaining, savepoint=savepoint)
 
     def __init__(self, dsn=None, user=None, password=None, host=None,
                     database=None, charset=DEFAULT_CHARSET, port=3050, 
@@ -662,6 +694,7 @@ class Connection(WireProtocol):
             self._op_attach()
         (h, oid, buf) = self._op_response()
         self.db_handle = h
+        self.closed = False
 
     def set_isolation_level(self, isolation_level):
         self.isolation_level = isolation_level
@@ -697,7 +730,7 @@ class Connection(WireProtocol):
         return r
 
     def _db_info_convert_type(self, info_request, v):
-        REQ_INT = [
+        REQ_INT = set([
             isc_info_allocation, isc_info_no_reserve, isc_info_db_sql_dialect,
             isc_info_ods_minor_version, isc_info_ods_version, 
             isc_info_page_size, isc_info_current_memory, isc_info_forced_writes,
@@ -709,13 +742,17 @@ class Connection(WireProtocol):
             isc_info_record_errors, isc_info_bpage_errors, 
             isc_info_dpage_errors, isc_info_ipage_errors,
             isc_info_ppage_errors, isc_info_tpage_errors,
-        ]
-        REQ_COUNT = [
+            # may not be available in some versions of Firebird
+            isc_info_oldest_transaction, isc_info_oldest_active, 
+            isc_info_oldest_snapshot, isc_info_next_transaction,
+            isc_info_active_tran_count
+        ])
+        REQ_COUNT = set([
             isc_info_backout_count, isc_info_delete_count, 
             isc_info_expunge_count, isc_info_insert_count, isc_info_purge_count,
             isc_info_read_idx_count, isc_info_read_seq_count, 
             isc_info_update_count
-        ]
+        ])
 
         if info_request in (isc_info_base_level, ):
             # IB6 API guide p52
@@ -748,6 +785,33 @@ class Connection(WireProtocol):
                 counts[bytes_to_int(v[i:i+2])] = bytes_to_int(v[i+2:i+6])
                 i += 6
             return counts
+        elif info_request in (isc_info_creation_date,):
+            nday = bytes_to_int(v[:4]) + 2400001 - 1721119
+            century = (4 * nday - 1) // 146097
+            nday = 4 * nday - 1 - 146097 * century
+            dd = nday // 4
+            nday = (4 * dd + 3) // 1461
+            dd = 4 * dd + 3 - 1461 * nday
+            dd = (dd + 4) // 4
+            mm = (5 * dd - 3) // 153
+            dd = 5 * dd - 3 - 153 * mm
+            dd = (dd + 5) // 5
+            yyyy = 100 * century + nday;
+            if mm < 10:
+                mm += 3
+            else:
+                mm -= 9
+                yyyy += 1
+
+            ntime = bytes_to_int(v[4:])
+            h = ntime // (3600 * ISC_TIME_SECONDS_PRECISION)
+            ntime %= 3600 * ISC_TIME_SECONDS_PRECISION
+            m = ntime // (60 * ISC_TIME_SECONDS_PRECISION)
+            ntime %= 60 * ISC_TIME_SECONDS_PRECISION
+            s = ntime // ISC_TIME_SECONDS_PRECISION
+            ms = ntime % ISC_TIME_SECONDS_PRECISION * 100
+
+            return datetime.datetime(yyyy, mm, dd, h, m, s, ms)
         else:
             return v
 
@@ -764,7 +828,7 @@ class Connection(WireProtocol):
             return results
 
     def close(self):
-        if not hasattr(self, "db_handle"):
+        if self.closed:
             return
         for trans in self._transactions:
             trans.close()
@@ -773,11 +837,16 @@ class Connection(WireProtocol):
         else:
             self._op_detach()
         (h, oid, buf) = self._op_response()
+        self.closed = True
+
+    def drop_database(self):
+        self._op_drop_database()
+        (h, oid, buf) = self._op_response()
+        self.closed = True
         delattr(self, "db_handle")
 
-
     def __del__(self):
-        if hasattr(self, "db_handle"):
+        if hasattr(self, "db_handle") and (not self.closed):
             self.close()
 
 class Transaction:
@@ -839,3 +908,62 @@ class Transaction:
     def closed(self):
         return not hasattr(self, "trans_handle")
 
+
+class RowMapping(Mapping):
+    """dict like interface to result rows
+    """
+    __slots__ = ("_description", "_fields")
+    
+    def __init__(self, row, description):
+        self._fields = fields = {}
+        # result may contain multiple fields with the same name. The
+        # RowMapping API ignores these additional fields.
+        for i, descr in enumerate(description):
+            fields.setdefault(descr[0], row[i])
+        self._description = description
+
+    def __getitem__(self, key):
+        fields = self._fields
+        # try unnormalized key first
+        try:
+            return fields[key]
+        except KeyError:
+            pass
+        
+        # normalize field name
+        if key[0] == '"' and key[-1] == '"':
+            # field names in quotes are case sensitive
+            normkey = key[1:-1]
+        else:
+            # default is all upper case fields
+            normkey = key.upper()
+        
+        try:
+            return fields[normkey]
+        except KeyError:
+            raise KeyError("RowMapping has no field names '%s'. Available "
+                           "field names are: %s" % 
+                           (key, ", ".join(self.keys())))
+            
+    def __iter__(self):
+        return iter(self._fields)
+    
+    def __len__(self):
+        return len(self._fields)
+    
+    def __repr__(self):
+        fields = self._fields
+        values = ["%s=%r" % (desc[0], fields[desc[0]]) 
+                  for desc in self._description]
+        return ("<RowMapping at 0x%08x with fields: %s>" % 
+                (id(self), ", ".join(values))) 
+
+    if not HAS_MAPPING:
+        def keys(self):
+            return list(self)
+        
+        def __setitem__(self, key, value):
+            raise NotImplementedError
+
+        def __delitem__(self, key):
+            raise NotImplementedError
