@@ -17,6 +17,8 @@ from firebirdsql.consts import *
 DEBUG = False
 
 if sys.version_info[0] == 2:
+    __metaclass__ = type
+
     def bytes(byte_array):
         return ''.join([chr(c) for c in byte_array])
 
@@ -46,30 +48,16 @@ def wire_operation(fn):
     return f
 
 def bytes_to_bint(b):           # Read as big endian
-    len_b = len(b)
-    if len_b == 1:
-        fmt = 'b'
-    elif len_b ==2:
-        fmt = '>h'
-    elif len_b ==4:
-        fmt = '>l'
-    elif len_b ==8:
-        fmt = '>q'
-    else:
+    fmtmap = {1: 'b', 2: '>h', 4: '>l', 8: '>q'}
+    fmt = fmtmap.get(len(b))
+    if fmt is None:
         raise InternalError
     return struct.unpack(fmt, b)[0]
 
 def bytes_to_int(b):            # Read as little endian.
-    len_b = len(b)
-    if len_b == 1:
-        fmt = 'b'
-    elif len_b ==2:
-        fmt = '<h'
-    elif len_b ==4:
-        fmt = '<l'
-    elif len_b ==8:
-        fmt = '<q'
-    else:
+    fmtmap = {1: 'b', 2: '<h', 4: '<l', 8: '<q'}
+    fmt = fmtmap.get(len(b))
+    if fmt is None:
         raise InternalError
     return struct.unpack(fmt, b)[0]
 
@@ -146,6 +134,7 @@ class WireProtocol:
     op_free_statement = 67
     op_prepare_statement = 68
     op_info_sql = 70
+    op_drop_database = 81
     op_service_attach = 82
     op_service_detach = 83
     op_service_info = 84
@@ -242,7 +231,7 @@ class WireProtocol:
         dpb += bytes([63, 4]) + int_to_bytes(3, 4) # isc_dpb_sql_dialect = 3
         dpb += bytes([24, 4]) + bint_to_bytes(1, 4) # isc_dpb_force_write = 1
         dpb += bytes([54, 4]) + bint_to_bytes(1, 4) # isc_dpb_overwirte = 1
-        dpb += bytes([4, 4]) + bint_to_bytes(page_size, 4)
+        dpb += bytes([4, 4]) + int_to_bytes(page_size, 4)
         p = xdrlib.Packer()
         p.pack_int(self.op_create)
         p.pack_int(0)                       # Database Object ID
@@ -279,6 +268,13 @@ class WireProtocol:
         p.pack_bytes(dpb)
         send_channel(self.sock, p.get_buffer())
 
+    @wire_operation
+    def _op_drop_database(self):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_drop_database)
+        p.pack_int(self.db_handle)
+        send_channel(self.sock, p.get_buffer())
+
     @wire_operation
     def _op_service_attach(self):
         dpb = bytes([2,2])
@@ -437,7 +433,7 @@ class WireProtocol:
         r += bint_to_bytes(3, 2)    # dialect
         if len(params) == 0:
             r += bint_to_bytes(0, 2)    # in_blr len
-            values = ''
+            values = bytes([])
         else:
             (blr, values) = self.params_to_blr(params)
             r += bint_to_bytes(len(blr), 2) + blr
